-- Time-stamp: <Thu Apr 08 2010 18:33:07 Stardate: Stardate: [-28]2888.65 hwloidl>
-- $Id: LUDecomp.hs,v 1.1.2.3 2002/06/15 01:34:28 hwloidl Exp $
--
-- Doing an LU decomposition on a square matrix represented as an Array.
-- The full details of this program, including a mathematical derivation
-- from its matrix algebra specification, can be found in:
-- 
-- Hudak, P., "Para-functional Programming in Haskell",
--   chapter 5 in "Parallel Functional Languages and Compilers",
--   edited by Szymanski, B.K., ACM Press (New York) and Addison-Wesley
--   (Reading), 1991.
--
-- Written by Paul Hudak, March 1993
--
-- Variants (CPP vars):
--  MOD_LU ... use modular arithmetic in LU decomposition and 
--             fwd/bwd_substitution to solve
--             top level function:
--             homsolv :: (Integral a) => 
--                        a -> Array (Int,Int) a -> Array (Int) a -> Array (Int) a
--  TRACE ... trace messages for debugging
-- -------------------------------------------------------------------------

module LUDecomp(lu, det, module Array, 
#ifdef MOD_LU
                homsolv,
#else
                linsolv_lu, 
#endif
                -- DEBUGGING ONLY        
                mult_mat_vec, mult_mat_mat, mult_vec_vec, 
                fwd_subst, bwd_subst, split_lu, check_lu, do_lu,
               ) where

import Array
import Ratio 
import Random
--import Strategies
#ifdef TRACE
--import Unsafe(trace)
import GlaExts(trace)
#endif
#ifdef MOD_LU
# if defined(WITH_BONZO)
import ModArithm hiding (modInv)
import qualified ModArithm (modInv)
# else
import ModArithm
# endif
#endif

-- bonzification

#if defined(WITH_BONZO)
modInv m x = if x==0 then 0 else ModArithm.modInv m x
#endif

-- Aux fcts

#if !defined(MOD_LU)
{-# SPECIALISE 
       sumOf :: (Int,Int) -> (Int -> Rational) -> Rational
  #-}
sumOf :: (Fractional a, Enum b) => (b,b) -> (b -> a) -> a
sumOf (a,b) f = foldl (+) 0.0 (map f [a..b])
#else
{-# SPECIALISE 
       sumOf :: Int -> (Int,Int) -> (Int -> Int) -> Int
  #-}
sumOf :: (Integral a, Enum b) => a -> (b,b) -> (b -> a) -> a
sumOf m (a,b) f = foldl (\ z x -> modSum m z x) 0 (map f [a..b])
#endif

{-
-- Hudak chapter
lu :: (Fractional a) => 
      Array (Int, Int) a -> Array (Int, Int) a
lu a = let 
         lua = array bds
               ( [ ((i,j), doL i j)  |  i<-[rl..rh], j<-[cl..i]   ] ++
                 [ ((i,j), doU i j)  |  i<-[rl..rh], j<-[i+1..ch] ] 
                 )
         doL i j = (a!(i,j) - sumOf (1,j-1) (\k -> lua!(i,k) * lua!(k,j)))
         doU i j 
           | lua!(i,i)==0 = error ("doU: 0 entry at (" ++ (show i) ++ "," ++ (show i) ++ ")")
           | otherwise = (a!(i,j) - sumOf (1,i-1) (\k -> lua!(i,k) * lua!(k,j))) 
                          / (lua!(i,i))
         bds@((rl,cl),(rh,ch)) = bounds a
       in 
       lua
-}

#if !defined(MOD_LU)

-- "Numerical Recipes", Chapter 2
{-# SPECIALISE 
  lu :: Array (Int,Int) Rational -> Array (Int,Int) Rational
  #-}
lu :: (Fractional a) => 
      Array (Int, Int) a -> Array (Int, Int) a
lu a = 
  let 
    lua = array bds
          ( [ ((i,j), doL i j)  |  j<-[cl..ch], i<-[j+1..rh] ] ++
            [ ((i,j), doU i j)  |  j<-[cl..ch], i<-[rl..j] ] 
            )
    -- i>j: (2.3.13)
    doL i j = (1 / lua!(j,j)) *  
              ((a!(i,j)) - sumOf (1,j-1) (\k -> (lua!(i,k)) * (lua!(k,j))))
    -- i<=j: (2.3.12)
    doU i j = (a!(i,j)) - sumOf (1,i-1) (\k -> (lua!(i,k)) * (lua!(k,j))) 
    bds@((rl,cl),(rh,ch)) = bounds a
  in 
  lua

fwd_subst :: (Fractional a) => 
             Array (Int,Int) a -> Array (Int) a -> Array (Int) a
fwd_subst lua b = 
  let
    y = array bds y'
    y' = ( [ (one, (b!1) / (lua!(one,one))) ] ++ 
           [ (i, 1 / (lua!(i,i)) * ((b!i) - sumOf (one,i-1) (\j -> (lua!(i,j))*(y!j))))
           | i <- [one+1..n] ] )
             
    bds@(one,n) = bounds b
    bds'@((rl,cl),(rh,ch)) = bounds lua
  in
#ifdef DEBUG
  if (one/=1 || rl/=1 || cl/=1 || n/=rh || n/=ch)
    then error $ "fwd_subst: Bounds error" ++ show bds ++ show bds'
    else 
#endif
#ifdef TRACE
         trace (show y') 
#endif 
         y

bwd_subst :: (Fractional a) => 
             Array (Int,Int) a -> Array (Int) a -> Array (Int) a
bwd_subst lua y =
  let
    x = array bds x'
    x' = ( [ (n, (y!n) / (lua!(n,n))) ] ++
           [ (i, 1 / (lua!(i,i)) * ((y!i) - sumOf (i+1,n) (\j -> (lua!(i,j))*(x!j))))
           | i <- [n-1,n-2..one] ] 
            )
    bds@(one,n) = bounds y
    bds'@((rl,cl),(rh,ch)) = bounds lua
  in
#ifdef DEBUG
  if (one/=1 || rl/=1 || cl/=1 || n/=rh || n/=ch)
    then error $ "bwd_subst: Bounds error" ++ show bds ++ show bds'
    else 
#endif
#ifdef TRACE
         trace (show x') 
#endif
         x

linsolv_lu :: (Fractional a) => 
              Array (Int,Int) a -> Array (Int) a -> Array (Int) a
linsolv_lu a b = 
  let
    lua = lu a
    (l,u) = split_lu lua
    y = fwd_subst l b
    x = bwd_subst u y
  in 
  x

det :: (Fractional a) => 
       Array (Int, Int) a -> a
det a = let
          bds@((rl,cl),(rh,ch)) = bounds a
          lua = lu a
        in 
        product [ lua!(j,j) | j <- [rl..rh] ]

#else /* MOD_LU ------------------------------------------------------------ */

-- same as above but using modular arithmetic
{-# SPECIALISE 
  lu :: Int -> Array (Int,Int) Int -> Array (Int,Int) Int 
  #-}
lu :: (Integral a) => 
      a -> Array (Int, Int) a -> Array (Int, Int) a
lu m a = 
  let 
    lua = array bds
          ( [ ((i,j), doL i j)  |  j<-[cl..ch], i<-[j+1..rh] ] ++
            [ ((i,j), doU i j)  |  j<-[cl..ch], i<-[rl..j] ] 
            )
    -- i>j: (2.3.13)
    doL i j =
#if defined(TRACE)
              let
                bonzo = if (lua!(j,j))==0
                         then 0
                         else modInv m (lua!(j,j))
              in   
              trace ("doL " ++ (show (i,j)) ++ (show bonzo)) $
#endif
              modProd m 
                (modInv m (lua!(j,j)))
                (modDif m (a!(i,j))
                          (sumOf m (1,j-1) (\k -> (modProd m (lua!(i,k)) 
                                                             (lua!(k,j))))))
    -- i<=j: (2.3.12)
    doU i j = 
#if defined(TRACE)
              trace ("doU " ++ (show (i,j))) $
#endif
              modDif m (a!(i,j)) 
                       (sumOf m (1,i-1) (\k -> modProd m (lua!(i,k)) (lua!(k,j)))) 
    bds@((rl,cl),(rh,ch)) = bounds a
  in 
  lua

{-# SPECIALISE 
  fwd_subst :: Int -> Array (Int,Int) Int -> Array (Int) Int -> Array (Int) Int
  #-}
fwd_subst :: (Integral a) => 
             a -> Array (Int,Int) a -> Array (Int) a -> Array (Int) a
fwd_subst m lua b = 
  let
    y = array bds y'
    y' = ( [ (one, modQuot m (b!1) (lua!(one,one))) ] ++ 
           [ (i, modProd m (modInv m (lua!(i,i)))
                           (modDif m (b!i) 
                                     (sumOf m (one,i-1) (\j -> modProd m (lua!(i,j)) (y!j)))))
           | i <- [one+1..n] ] )
             
    bds@(one,n) = bounds b
    bds'@((rl,cl),(rh,ch)) = bounds lua
  in
#ifdef DEBUG
  if (one/=1 || rl/=1 || cl/=1 || n/=rh || n/=ch)
    then error $ "fwd_subst: Bounds error" ++ show bds ++ show bds'
    else 
#endif
#ifdef TRACE
         trace (show y') 
#endif 
         y

{-# SPECIALISE 
  bwd_subst :: Int -> Array (Int,Int) Int -> Array (Int) Int -> Array (Int) Int
  #-}
bwd_subst :: (Integral a) => 
             a -> Array (Int,Int) a -> Array (Int) a -> Array (Int) a
bwd_subst m lua y =
  let
    x = array bds x'
    x' = ( [ (n, modQuot m (y!n) (lua!(n,n))) ] ++
           [ (i, (modProd m (modInv m (lua!(i,i)))
                            (modDif m (y!i) 
                                      (sumOf m (i+1,n) (\j -> modProd m (lua!(i,j)) (x!j))))))
           | i <- [n-1,n-2..one] ] 
            )
    bds@(one,n) = bounds y
    bds'@((rl,cl),(rh,ch)) = bounds lua
  in
#ifdef DEBUG
  if (one/=1 || rl/=1 || cl/=1 || n/=rh || n/=ch)
    then error $ "bwd_subst: Bounds error" ++ show bds ++ show bds'
    else 
#endif
#ifdef TRACE
         trace (show x') 
#endif
         x

{-# SPECIALISE 
  homsolv :: Int -> Array (Int,Int) Int -> Array (Int) Int -> Array (Int) Int
  #-}
homsolv :: (Integral a) => 
           a -> Array (Int,Int) a -> Array (Int) a -> Array (Int) a
homsolv m a b = 
  let
    lua = lu m a
    (l,u) = split_lu m lua
    y = fwd_subst m l b
    x = bwd_subst m u y
  in 
  x

{-# SPECIALISE 
  det :: Int -> Array (Int, Int) Int -> Int
  #-}
det :: (Integral a) => 
       a -> Array (Int, Int) a -> a
det m a = let
          bds@((rl,cl),(rh,ch)) = bounds a
          lua = lu m a
        in 
        foldl (\ z x -> modProd m z x) 0 [ lua!(j,j) | j <- [rl..rh] ]

#endif

-- debugging only

#if !defined(MOD_LU)

mult_mat_vec :: (Fractional a) => 
                Array (Int,Int) a -> Array (Int) a -> Array (Int) a
mult_mat_vec m v = array (bounds v) 
                     [ (i, sum [ (m!(i,j))*(v!j) | j <- [cl..ch] ])
                     | i <- [rl..rh] ]
                   where bds@((rl,cl),(rh,ch)) = bounds m
        
mult_mat_mat :: (Fractional a) => 
                Array (Int,Int) a -> Array (Int,Int) a -> Array (Int,Int) a
mult_mat_mat m1 m2= array (bounds m1) 
                     [ ((i,j), sum [ (m1!(i,k))*(m2!(k,j)) | k <- [rl..rh] ] ) 
                     | j <- [cl..ch] , i <- [rl..rh] ]
                   where bds@((rl,cl),(rh,ch)) = bounds m1

mult_vec_vec v1 v2 = array bds (sum [ v1!i * v2!i | i<-[l..h] ])
                     where bds@(l,h) = bounds v1

split_lu :: (Fractional a) => 
            Array (Int,Int) a -> (Array (Int,Int) a, Array (Int,Int) a)
split_lu lu = 
  let 
    l = array bnds 
         ( [ ((i,j), 0) | i <- [rl..rh], j <- [cl..ch] , i<j ] ++
           [ ((i,i), 1) | i <- [rl..rh] ] ++
           [ ((i,j), lu!(i,j)) | i <- [rl..rh], j <- [cl..ch] , i>j ])
    u = array bnds 
         ( [ ((i,j), lu!(i,j)) | i <- [rl..rh], j <- [cl..ch] , i<=j ] ++
           [ ((i,j), 0) | i <- [rl..rh], j <- [cl..ch] , i>j ])
    bnds@((rl,cl),(rh,ch)) = bounds lu
  in
  (l,u)

check_lu :: (Fractional a) => 
            Array (Int,Int) a -> Array (Int,Int) a -> Bool
check_lu lu m = 
  let
    (l,u) = split_lu m
  in
  if (mult_mat_mat l u) == m
    then True
    else False

do_lu :: (Fractional a) => Array (Int,Int) a -> IO ()
do_lu m = do
            let x = lu m
            let (l,u) = split_lu x
            if (mult_mat_mat l u)==m
              then putStrLn "OK"
              else putStrLn "WRONG"

#else

mult_mat_vec :: (Integral a) => 
                a -> Array (Int,Int) a -> Array (Int) a -> Array (Int) a
mult_mat_vec q m v = array (bounds v) 
                     [ (i, foldl (\ z x -> modSum q z x) 0 [ (m!(i,j))*(v!j) | j <- [cl..ch] ])
                     | i <- [rl..rh] ]
                   where bds@((rl,cl),(rh,ch)) = bounds m
        
mult_mat_mat :: (Integral a) => 
                a -> Array (Int,Int) a -> Array (Int,Int) a -> Array (Int,Int) a
mult_mat_mat q m1 m2= array (bounds m1) 
                     [ ((i,j), foldl (\ z x -> modSum q z x) 0  [ modProd q (m1!(i,k)) (m2!(k,j)) | k <- [rl..rh] ] ) 
                     | j <- [cl..ch] , i <- [rl..rh] ]
                   where bds@((rl,cl),(rh,ch)) = bounds m1

mult_vec_vec q v1 v2 = array bds (foldl (\ z x -> modSum q z x) 0 [ modProd q (v1!i)  (v2!i) | i<-[l..h] ])
                     where bds@(l,h) = bounds v1

split_lu :: (Integral a) => 
            a -> Array (Int,Int) a -> (Array (Int,Int) a, Array (Int,Int) a)
split_lu _ lu = 
  let 
    l = array bnds 
         ( [ ((i,j), 0) | i <- [rl..rh], j <- [cl..ch] , i<j ] ++
           [ ((i,i), 1) | i <- [rl..rh] ] ++
           [ ((i,j), lu!(i,j)) | i <- [rl..rh], j <- [cl..ch] , i>j ])
    u = array bnds 
         ( [ ((i,j), lu!(i,j)) | i <- [rl..rh], j <- [cl..ch] , i<=j ] ++
           [ ((i,j), 0) | i <- [rl..rh], j <- [cl..ch] , i>j ])
    bnds@((rl,cl),(rh,ch)) = bounds lu
  in
  (l,u)

check_lu :: (Integral a) => 
            a -> Array (Int,Int) a -> Array (Int,Int) a -> Bool
check_lu q lu m = 
  let
    (l,u) = split_lu q m
  in
  if (mult_mat_mat q l u) == m
    then True
    else False

do_lu :: (Integral a) => a -> Array (Int,Int) a -> IO ()
do_lu q m = do
            let x = lu q m
            let (l,u) = split_lu q x
            if (mult_mat_mat q l u)==m
              then putStrLn "OK"
              else putStrLn "WRONG"

#endif

            
-- ---------------------------------------------------------------------------

-- a1 :: Array (Int, Int) Float
-- a1 = array ((1,1),(3,3)) 
--        [((1,1) ,  1.0), ((1,2) ,  2.0), ((1,3) ,  3.0), 
--         ((2,1) ,  4.0), ((2,2) , -3.0), ((2,3) ,  2.0), 
-- 	((3,1) ,  7.0), ((3,2) , -6.0), ((3,3) ,  0.0)]
-- -- det a1 = 31

a11 :: Array (Int, Int) (Ratio Integer)
a11 = array ((1,1),(3,3)) 
       [((1,1) ,  1 % 2), ((1,2) ,  2 % 3), ((1,3) ,  3 % 2), 
        ((2,1) ,  4 % 1), ((2,2) , -3 % 2), ((2,3) ,  2 % 5), 
	((3,1) ,  7 % 2), ((3,2) , -6 % 5), ((3,3) ,  0 % 1)]
{- in calc.el format
       [ [1:2,  2:3, 3:2], 
         [4:1, -3:2, 2:5], 
	 [7:2, -6:5, 0:1] ]
    det: 1109:600
-}

listToListList :: Int -> [a] -> [[a]]
listToListList c m 
 | length m <= c = [m]
 | otherwise = c1 : listToListList c resto
  where (c1,resto) = splitAt c m

lolSqMatrix bds@((xlo,ylo), (xhi, yhi)) zss = 
  array bds (concat [ [ ((x,y), (zss!!(x-1))!!(y-1)) | y <- [ylo..yhi] ] | x <- [xlo..xhi] ])
            
lolArr bds@((xlo,ylo), (xhi, yhi)) zss = 
  array bds (concat [ [ ((x,y), (zss!!(x-1))!!(y-1)) | y <- [ylo..yhi] ] 
                    | x <- [xlo..xhi] ])

