{-# LANGUAGE BangPatterns #-}

-----------------------------------------------------------------------------
-- |
-- Module      :  Control.Parallel.Strategies
-- Copyright   :  (c) The University of Glasgow 2001-2009
-- License     :  BSD-style (see the file libraries/base/LICENSE)
--
-- Maintainer  :  libraries@haskell.org
-- Stability   :  experimental
-- Portability :  portable
--
-- Parallel Evaluation Strategies, or Strategies for short, provide
-- ways to express parallel computations.  Strategies have the following
-- key features:
--
--  * Strategies express /deterministic parallelism/:
--    the result of the program is unaffected by evaluating in parallel.
--    The parallel tasks evaluated by a Strategy may have no side effects.
--    For non-deterministic parallel programming, see "Control.Concurrent".
--
--  * Strategies let you separate the description of the parallelism from the
--    logic of your program, enabling modular parallelism.  The basic idea
--    is to build a lazy data structure representing the computation, and
--    then write a Strategy that describes how to traverse the data structure
--    and evaluate components of it sequentially or in parallel.
--
--  * Strategies are /compositional/: larger strategies can be built
--    by gluing together smaller ones.
--
--  * 'Monad' and 'Applicative' instances are provided, for quickly building
--    strategies that involve traversing structures in a regular way.
-- 
-- The strategies library has a long history.  What follows is a
-- summary of how the current design evolved, and is mostly of
-- interest to those who are familiar with an older version, or need
-- to adapt old code to use the newer API.
--
-- Version 1.x
--
--   The original Strategies design is described in /Algorithm + Strategy = Parallelism/ <http://www.macs.hw.ac.uk/~dsg/gph/papers/html/Strategies/strategies.html>
--   and the code was written by
--      Phil Trinder, Hans-Wolfgang Loidl, Kevin Hammond et al. 
--
-- Version 2.x
--
-- Later, during work on the shared-memory implementation of
-- parallelism in GHC, we discovered that the original formulation of
-- Strategies had some problems, in particular it lead to space leaks
-- and difficulties expressing speculative parallelism.  Details are in
-- the paper /Runtime Support for Multicore Haskell/ <http://www.haskell.org/~simonmar/papers/multicore-ghc.pdf>.
--
-- This module has been rewritten in version 2. The main change is to
-- the 'Strategy a' type synonym, which was previously @a -> Done@ and
-- is now @a -> Eval a@.  This change helps to fix the space leak described
-- in \"Runtime Support for Multicore Haskell\".  The problem is that
-- the runtime will currently retain the memory referenced by all
-- sparks, until they are evaluated.  Hence, we must arrange to
-- evaluate all the sparks eventually, just in case they aren't
-- evaluated in parallel, so that they don't cause a space leak.  This
-- is why we must return a \"new\" value after applying a 'Strategy',
-- so that the application can evaluate each spark created by the
-- 'Strategy'.
--
-- The simple rule is this: you /must/ use the result of applying
-- a 'Strategy' if the strategy creates parallel sparks, and you
-- should probably discard the the original value.  If you don't
-- do this, currently it may result in a space leak.  In the
-- future (GHC 6.14), it will probably result in lost parallelism
-- instead, as we plan to change GHC so that unreferenced sparks
-- are discarded rather than retained (we can't make this change
-- until most code is switched over to this new version of
-- Strategies, because code using the old verison of Strategies
-- would be broken by the change in policy).
--
-- The other changes in version 2.x are:
--
--   * Strategies can now be defined using a convenient Monad/Applicative
--     type, 'Eval'.  e.g. @parList s = traverse (rpar `dot` s)@
--
--   * 'parList' has been generalised to 'parTraversable', which works on
--     any 'Traversable' type, and similarly 'seqList' has been generalised
--     to 'evalTraversable'.
--
--   * 'parList' and 'parBuffer' have versions specialised to 'rseq',
--     and there are transformation rules that automatically translate
--     e.g. @parList rseq@ into a call to the optimised version.
--
--   * 'NFData' has been moved to @Control.DeepSeq@ in the @deepseq@
--     package.  Note that since the 'Strategy' type changed, 'rnf'
--     is no longer a 'Strategy': use 'rdeepseq' instead.
--
--   * 'rwhnf' has been renamed to 'rseq' to reflect its relationship
--     with 'rdeepseq'.
--
-- Version 2.1 moved 'NFData' into a separate package, @deepseq@.
--
-- Version 2.2 changed the type of Strategy to @a -> Eval a@, and
-- re-introduced the @r0@ strategy which was missing in version 2.1.
--
-- Version 2.3 simplified the @Eval@ type, so that @Eval@ is now just
-- the strict identity monad.  This change and various other
-- improvements and refactorings are thanks to Patrick Maier who
-- noticed that @Eval@ didn't satisfy the monad laws, and that a
-- simpler version would fix that problem.
--
-- Version 2.4 introduced sequential strategies (for controlling evaluation) 
-- as a subtype of strategies. Other refactorings include the following
-- strategy naming scheme:
--
--   * Basic polymorphic strategies (of type @'Strategy' a@) are called @r...@.
--     Examples: 'r0', 'rseq', 'rpar', 'rdeepseq'.
--
--   * A strategy combinator for a particular type constructor 
--     or constructor class @T@ is called @evalT...@, @parT...@ or @seqT...@.
--
--   ** The @seqT...@ combinators (residing in module
--      "Control.Parallel.SeqStrategies") yield sequential strategies.
--      Thus, @seqT...@ combinators cannot spark, nor can the sequential
--      strategies to which they may be applied.
--      Examples: 'seqTuple2', 'seqListN', 'seqFoldable'.
--
--   ** The @evalT...@ combinators do not spark themselves, yet they may
--      be applied to strategies that do spark. (They may also be applied
--      to non-sparking strategies; however, in that case the corresponding
--      @seqT...@ combinator might be a better choice.)
--      Examples: 'evalTuple2', 'evalListN', 'evalTraversable'.
--
--   ** The @parT...@ combinators, which are derived from their @evalT...@
--      counterparts, do spark. They may be applied to all strategies,
--      whether sparking or not.
--      Examples: 'parTuple2', 'parListN', 'parTraversable'.
--
--   ** An exception to the type driven naming scheme are 'evalBuffer' and
--      'parBuffer', which are not named after their type constructor (lists)
--      but after their function (rolling buffer of fixed size).
--
--   * A strategy combinator that is not as compositional as its type
--     suggests is suffixed with @'@.
--     Examples: 'evalFunctor'', 'parBuffer''.
--

module Control.Parallel.Strategies
       ( -- * User interface
         -- ** The strategy type
         Strategy

       , strategies_version -- version information

         -- ** Injection of sequential strategies
       , ins               -- :: Control.Parallel.SeqStrategies.Strategy a -> Strategy a

         -- ** Application of strategies
       , using             -- :: a -> Strategy a -> a
       , withStrategy      -- :: Strategy a -> a -> a

         -- ** Composition of strategies
       , dot               -- :: Strategy a -> Strategy a -> Strategy a

         -- ** Basic strategies
       , r0                -- :: Strategy a
       , rseq
       , rdeepseq          -- :: NFData a => Strategy a
       , rpar              -- :: Strategy a

         -- ** Strategies for traversable data types
       , evalTraversable   -- :: Traversable t => Strategy a -> Strategy (t a)
       , parTraversable
       , evalMonad         -- :: Monad m => Strategy a -> Strategy (m a)
       , parMonad
       , evalFunctor'      -- :: Functor f => Strategy a -> Strategy (f a) -> Strategy (f a)
       , parFunctor'

         -- ** Strategies for lists
       , evalList          -- :: Strategy a -> Strategy [a]
       , parList
       , evalListN         -- :: Int -> Strategy a -> Strategy [a]
       , parListN
       , evalListNth       -- :: Int -> Strategy a -> Strategy [a]
       , parListNth
       , evalListSplitAt   -- :: Int -> Strategy [a] -> Strategy [a] -> Strategy [a]
       , parListSplitAt

         -- ** Strategies for lazy lists
       , evalBuffer        -- :: Int -> Strategy a -> Strategy [a]
       , parBuffer
       , evalBuffer'       -- :: Int -> Strategy a -> Strategy [a]
       , parBuffer'

         -- ** Strategies for tuples
       , evalTuple2        -- :: Strategy a -> ... -> Strategy (a,...)
       , evalTuple3
       , evalTuple4
       , evalTuple5
       , evalTuple6
       , evalTuple7
       , evalTuple8
       , evalTuple9
       , parTuple2         -- :: Strategy a -> ... -> Strategy (a,...)
       , parTuple3
       , parTuple4
       , parTuple5
       , parTuple6
       , parTuple7
       , parTuple8
       , parTuple9

         -- ** Strategic function application
       , ($|)              -- :: (a -> b) -> Strategy a -> a -> b
       , ($||)
       , (.|)              -- :: (b -> c) -> Strategy b -> (a -> b) -> a -> c
       , (.||)
       , (-|)              -- :: (a -> b) -> Strategy b -> (b -> c) -> a -> c
       , (-||)

         -- * Hackers' interface (for strategy programmers)
         -- ** The Eval monad
       , Eval(Done)        -- instances: Monad, Functor, Applicative
       , runEval           -- :: Eval a -> a

         -- * Deprecated functionality
    -- , demanding, sparking   -- All variants of par and pseq banned.
    -- , (>|), (>||)           -- Use rpar and rseq instead.
       ) where

import Prelude
import Control.Parallel (par, pseq)
import Control.DeepSeq (NFData, rnf)
import Control.Applicative (Applicative, pure, (<*>))
import Control.Monad (liftM, ap)
import Data.Traversable (Traversable, traverse)
import Data.Foldable (toList)
import Data.Sequence (Seq, (|>), fromList, ViewL((:<)), viewl)
import qualified Control.Parallel.SeqStrategies as SeqStrategies
-- import qualified SeqStrategies

infixr 9 `dot`     -- same as function composition (.)
infixl 0 `using`   -- lowest precedence and associate to the left

-- Version (should correspond to Cabal's meta-info on the package)
strategies_version = "3.0.0.4"

-- --------------------------------------------------------------------------
-- Eval monad (isomorphic to Lift monad from MonadLib 3.6.1)

-- | 'Eval' is a Monad that makes it easier to define parallel
-- strategies.  It is a strict identity monad: that is, in 
--
--  > m >>= f
--
-- @m@ is evaluated before the result is passed to @f@.
--
--  > instance Monad Eval where
--  >   return  = Done
--  >   m >>= k = case m of
--  >               Done x -> k x
--
-- If you wanted to construct a 'Strategy' for a pair that sparked the
-- first component in parallel and then evaluated the second
-- component, you could write
--
-- > myStrat :: Strategy (a,b)
-- > myStrat (a,b) = do { a' <- rpar a; b' <- rseq b; return (a',b') }
--
-- Alternatively, you could write this more compactly using the
-- Applicative style as
--
-- > myStrat (a,b) = (,) <$> rpar a <*> rseq b
--
-- More examples, using the Applicative instance:
--
-- > parList :: Strategy a -> Strategy [a]
-- > parList strat = traverse (rpar `dot` strat)
--
-- > evalPair :: Strategy a -> Strategy b -> Strategy (a,b)
-- > evalPair f g (a,b) = pure (,) <$> f a <*> g b
--

data Eval a = Done a

-- | Pull the result out of the monad.
runEval :: Eval a -> a
runEval (Done x) = x

instance Monad Eval where
  return x = Done x
  Done x >>= k = k x   -- Note: pattern 'Done x' makes '>>=' strict

instance Functor Eval where
  fmap = liftM

instance Applicative Eval where
  (<*>) = ap
  pure  = return


-- --------------------------------------------------------------------------
-- Strategies

-- | A 'Strategy' is a function that embodies a parallel evaluation strategy.
-- The function traverses (parts of) its argument, evaluating subexpressions
-- in parallel or in sequence.
-- 
-- A 'Strategy' may do an arbitrary amount of evaluation of its
-- argument, but should not return a value different from the one it
-- was passed.
--
-- Parallel computations may be discarded by the runtime system if the
-- program no longer requires their result, which is why a 'Strategy'
-- function returns a new value equivalent to the old value.  The
-- intention is that the program applies the 'Strategy' to a
-- structure, and then uses the returned value, discarding the old
-- value.  This idiom is expressed by the 'using' function.
-- 
type Strategy a = a -> Eval a

-- | Evaluate a value using the given strategy.
--
-- > x `using` s = runEval (s x)
--
using :: a -> Strategy a -> a
x `using` strat = runEval (strat x)

-- Apparently code that does not allocate memory may suffer in the
-- uptake of parallelism because the GHC scheduler (responsible for
-- globalising sparks) gets triggered only by memory allocation. To
-- circumvent this problem, we could provide a version of 'using' that
-- deliberately triggers the scheduler before returning by allocating a
-- little garbage.  An example is the function 'usingS', which wraps the
-- result of applying the strategy in an 'Either' type and then unwraps
-- it again.  Of cource, the compiler must be stopped from optimising
-- away the unnecessary wrapping.
--
-- usingS :: a -> Strategy a -> a
-- x `usingS` strat = either id id (runEval (strat x >>= return . Left))
--
-- If this approach is adopted then 'usingS' should become the normal
-- user interface, whereas the plain 'using' should be demoted to the
-- hackers' interface. It may be used to build strategies by applying
-- strategies but it should never be used as the top-level strategy
-- application function.

-- | Evaluate a value using the given strategy. 
-- This is simply 'using' with arguments reversed.
withStrategy :: Strategy a -> a -> a
withStrategy = flip using

-- | Compose two strategies sequentially. 
-- This is the analogue to function composition on strategies.
--
-- > strat2 `dot` strat1 == strat2 . withStrategy strat1
--
dot :: Strategy a -> Strategy a -> Strategy a
strat2 `dot` strat1 = strat2 . runEval . strat1

-- Proof of strat2 `dot` strat1 == strat2 . withStrategy strat1
--
--    strat2 . withStrategy strat1
-- == \x -> strat2 (withStrategy strat1 x)
-- == \x -> strat2 (x `using` strat1)
-- == \x -> strat2 (runEval (strat1 x))
-- == \x -> (strat2 . runEval . strat1) x
-- == strat2 `dot` strat1

-- One might be tempted to think that 'dot' is equivalent to '(<=<)',
-- the right-to-left Kleisli composition in the Eval monad, because
-- '(<=<)' can take the type @Strategy a -> Strategy a -> Strategy a@
-- and intuitively does what 'dot' does: First apply the strategy to the
-- right then the one to the left. However, there is a subtle difference
-- in strictness, witnessed by the following example:
--
-- > (r0 `dot` rseq) undefined == Done undefined
-- > (r0 <=< rseq) undefined == undefined
--


-- | Type synonym to improve readability of the type of 'ins'.
-- type SeqStrategy a = Control.Parallel.SeqStrategies.Strategy a
type SeqStrategy a = SeqStrategies.Strategy a

-- | Inject a sequential strategy (ie. coerce a sequential strategy
-- to a general strategy).
--
-- Thanks to 'ins', the type @SeqStrategy a@ (which is a synonym for
-- @"Control.Parallel.SeqStrategies".Strategy a@) is a subtype
-- of @'Strategy' a@.
ins :: SeqStrategy a -> Strategy a
ins strat x = strat x `pseq` return x


-- --------------------------------------------------------------------------
-- Basic strategies (some imported from SeqStrategies)

-- | 'r0' performs *no* evaluation.
--
-- > r0 == ins Control.Parallel.SeqStrategies.r0
--
r0 :: Strategy a
r0 x = return x

-- Proof of r0 == ins SeqStrategies.r0
--
--    ins SeqStrategies.r0
-- == \x -> SeqStrategies.r0 x `pseq` return x
-- == \x -> SeqStrategies.Done `pseq` return x
-- == \x -> return x
-- == r0

-- | 'rseq' evaluates its argument to weak head normal form.
--
-- > rseq == ins Control.Parallel.SeqStrategies.rseq
--
rseq :: Strategy a
rseq x = x `pseq` return x

-- Proof of rseq == ins SeqStrategies.rseq
--
--    ins SeqStrategies.rseq
-- == \x -> SeqStrategies.rseq x `pseq` return x
-- == \x -> (x `seq` SeqStrategies.Done) `pseq` return x
-- == \x -> x `pseq` return x
-- == rseq

-- | 'rdeepseq' fully evaluates its argument.
--
-- > rdeepseq == ins Control.Parallel.SeqStrategies.rdeepseq
--
rdeepseq :: NFData a => Strategy a
rdeepseq x = rnf x `pseq` return x

-- Proof of rdeepseq == ins SeqStrategies.rdeepseq
--
--    ins SeqStrategies.rdeepseq
-- == \x -> SeqStrategies.rdeepseq x `pseq` return x
-- == \x -> (x `deepseq` SeqStrategies.Done) `pseq` return x
-- == \x -> (rnf x `seq` SeqStrategies.Done) `pseq` return x
-- == \x -> rnf x `pseq` return x
-- == rdeepseq

-- | 'rpar' sparks its argument (for evaluation in parallel).
rpar :: Strategy a
rpar x = x `par` return x


-- --------------------------------------------------------------------------
-- Strategy combinators for Traversable data types

-- | Evaluate the elements of a traversable data structure 
-- according to the given strategy.
{-# INLINE evalTraversable #-}
evalTraversable :: Traversable t => Strategy a -> Strategy (t a)
evalTraversable = traverse

-- | Like 'evalTraversable' but evaluates all elements in parallel.
{-# INLINE parTraversable #-}
parTraversable :: Traversable t => Strategy a -> Strategy (t a)
parTraversable strat = evalTraversable (rpar `dot` strat)

{-# SPECIALISE evalTraversable :: Strategy a -> Strategy (Maybe a) #-}
{-# SPECIALISE parTraversable :: Strategy a -> Strategy (Maybe a) #-}
{-# SPECIALISE evalTraversable :: Strategy a -> Strategy [a] #-}
{-# SPECIALISE parTraversable :: Strategy a -> Strategy [a] #-}


-- --------------------------------------------------------------------------
-- Strategy combinators for Monads

-- | Evaluate a monad by lifting the given strategy for the base type.
--
-- For Traversable functors which are Monads (eg. []), the types of
-- 'evalTraversable' and 'evalMonad' overlap. The two functions are not
-- necessarily the same, though. There is a difference in strictness,
-- as witnessed by the following example:
--
-- > take 5 ([1..] `using` evalTraversable r0) == undefined
-- > take 5 ([1..] `using` evalMonad r0) == [1,2,3,4,5]
--
evalMonad :: Monad m => Strategy a -> Strategy (m a)
evalMonad strat = return . liftM (withStrategy strat)

-- | Like 'evalMonad' but evaluates the lifting in parallel.
parMonad :: Monad m => Strategy a -> Strategy (m a)
parMonad strat = evalMonad (rpar `dot` strat)


-- --------------------------------------------------------------------------
-- Strategy combinators based on (potentially lazy) Functor strategies

-- | @'evalFunctor'' strat shape@ evaluates a functorial data structure
-- according to the strategies @shape@ (which is supposed to control the
-- evaluaton of the shape of the data structure) and @strat@ (which is
-- supposed to control the evaluation of the elements).
--
-- 'evalFunctor'' is not as compositional as its type suggests. In fact,
-- the strategy @shape@ dominates @strat@ in the sense that @strat@ may be
-- disregarded depending on shape, as in following examples for lists.
--
-- > evalFunctor' strat r0           == r0
-- > evalFunctor' strat (seqList r0) == seqList r0
-- > evalFunctor' strat rdeepseq     == rdeepseq
--
-- The reason is that @strat@ only controls evaluation of elements beyond
-- weak head normal form, and only for elements that are evaluated (to
-- weak head normal form at least) by @shape@.
evalFunctor' :: Functor f => Strategy a -> Strategy (f a) -> Strategy (f a)
evalFunctor' strat shape x = return (fmap (withStrategy strat) x `using` shape)

-- | Like 'evalFunctor'' but evaluates all elements in parallel.
parFunctor' :: Functor f => Strategy a -> Strategy (f a) -> Strategy (f a)
parFunctor' strat = evalFunctor' (rpar `dot` strat)

{-# SPECIALISE evalFunctor' :: Strategy a -> Strategy [a] -> Strategy [a] #-}
{-# SPECIALISE parFunctor' :: Strategy a -> Strategy [a] -> Strategy [a] #-}


-- --------------------------------------------------------------------------
-- Strategies for lists

-- | Evaluate each element of a list according to the given strategy.
--  Equivalent to 'evalTraversable' at the list type.
evalList :: Strategy a -> Strategy [a]
evalList = evalTraversable
-- Alternative explicitly recursive definition:
-- evalList strat []     = return []
-- evalList strat (x:xs) = strat x >>= \x' ->
--                         evalList strat xs >>= \xs' ->
--                         return (x':xs')

-- | Evaluate each element of a list in parallel according to given strategy.
--  Equivalent to 'parTraversable' at the list type.
parList :: Strategy a -> Strategy [a]
parList = parTraversable
-- Alternative definition via evalList:
-- parList strat = evalList (rpar `dot` strat)

-- | @'evaListSplitAt' n stratPref stratSuff@ evaluates the prefix
-- (of length @n@) of a list according to @stratPref@ and its the suffix
-- according to @stratSuff@.
evalListSplitAt :: Int -> Strategy [a] -> Strategy [a] -> Strategy [a]
evalListSplitAt n stratPref stratSuff xs
  = let (ys,zs) = splitAt n xs in
    stratPref ys >>= \ys' ->
    stratSuff zs >>= \zs' ->
    return (ys' ++ zs')

-- | Like 'evalListSplitAt' but evaluates both sublists in parallel.
parListSplitAt :: Int -> Strategy [a] -> Strategy [a] -> Strategy [a]
parListSplitAt n stratPref stratSuff = evalListSplitAt n (rpar `dot` stratPref) (rpar `dot` stratSuff)

-- | Evaluate the first n elements of a list according to the given strategy.
evalListN :: Int -> Strategy a -> Strategy [a]
evalListN n strat = evalListSplitAt n (evalList strat) r0

-- | Like 'evalListN' but evaluates the first n elements in parallel.
parListN :: Int -> Strategy a -> Strategy [a]
parListN n strat = evalListN n (rpar `dot` strat)

-- | Evaluate the nth element of a list (if there is such) according to
-- the given strategy.
-- The spine of the list up to the nth element is evaluated as a side effect.
evalListNth :: Int -> Strategy a -> Strategy [a]
evalListNth n strat = evalListSplitAt n r0 (evalListN 1 strat)

-- | Like 'evalListN' but evaluates the nth element in parallel.
parListNth :: Int -> Strategy a -> Strategy [a]
parListNth n strat = evalListNth n (rpar `dot` strat)

-- Non-compositional version of 'parList', evaluating list elements
-- to weak head normal form.
-- Not to be exported; used for optimisation.
parListWHNF :: Strategy [a]
parListWHNF xs = go xs `pseq` return xs
  where -- go :: [a] -> [a]
           go []     = []
           go (y:ys) = y `par` go ys

-- The non-compositional 'parListWHNF' might be more efficient than its
-- more compositional counterpart; use RULES to do the specialisation.

{-# RULES 
 "parList/rseq" parList rseq = parListWHNF
 #-}


-- --------------------------------------------------------------------------
-- Strategies for lazy lists

-- | 'evalBuffer' is a rolling buffer strategy combinator for (lazy) lists.
-- It maintains an n place FIFO buffer of elements already evaluated (or
-- under evaluation) according to the given strategy. Every time the head of
-- the buffer is demanded, a new element is evaluated and enqueued at the tail.
--
-- To illustrate the semantics we exhibit the simple (yet inefficient, for
-- large n) list-based implementation below.
-- 
-- > evalBuffer n strat xs = return (roll (xs `using` evalListN n strat))
-- >   where roll []     = []
-- >         roll (x:xs) = x : roll (xs `using` evalListNth n strat)
--
-- The actual implementation realises the buffer as a functional queue
-- (with amortised constant-time operations).
evalBuffer :: Int -> Strategy a -> Strategy [a]
evalBuffer n strat xs = roll (fromList (ys `using` evalList strat)) zs
  where (ys,zs) = splitAt n xs

     -- roll :: Seq a -> [a] -> Eval [a]    -- Note: return type is *not* Strategy [a] as it *not* an identity
        roll q []     = return (toList q)
        roll q (z:zs) = do z' <- strat z
                           let y':<q' = viewl (q |> z') -- Note: Match y':<q' cannot fail because q|>z' is non-empty.                           
                           return (y' : runEval (roll q' zs))

-- | Like 'evalBuffer' but evaluates the list elements in parallel when
-- pushing them into the buffer.
parBuffer :: Int -> Strategy a -> Strategy [a]
parBuffer n strat = evalBuffer n (rpar `dot` strat)

{-
-- Another rolling buffer strategy combinator.
-- This one realises the buffer as a DiffArray.
-- In principle, this should yield buffer strategies that are worst case
-- constant space and constant time.
-- However, the DiffArray module has currently performance problems,
-- which translate into poor performance of this buffer strategy.
--
-- This code requires:
--
-- > {-# LANGUAGE ScopedTypeVariables #-}
-- > import Data.Array.Diff (DiffArray, listArray, (!), (//))
--
evalBufferA :: forall a. Int -> Strategy a -> Strategy [a]
evalBufferA n strat = return . roll . init
  where init :: [a] -> (Int, DiffArray Int (Maybe a), [Maybe a])
        init xs = (0, listArray (0,n-1) ys', zs')
          where (ys,zs) = splitAt n xs
                ys' = map Just (ys `using` evalList strat) ++ repeat Nothing
                zs' = map Just zs ++ repeat Nothing

        roll :: (Int, DiffArray Int (Maybe a), [Maybe a]) -> [a]
        roll buf = case rotate buf of
                     (Nothing, _)    -> []
                     (Just x,  buf') -> x : roll buf'

        rotate :: (Int, DiffArray Int (Maybe a), [Maybe a]) -> (Maybe a, (Int, DiffArray Int (Maybe a), [Maybe a]))
        rotate (i, arr, feed) = (arr!i, ((i+1) `mod` n, arr // [(i, z `using` evalTraversable strat)], zs))
          where z:zs = feed
-}

-- List-based non-compositional rolling buffer strategy, evaluating list
-- elements to weak head normal form.
-- Not to be exported; used in evalBuffer' and for optimisation.
evalBufferWHNF :: Int -> Strategy [a]
evalBufferWHNF n xs = return (ret xs (start n xs))
  where -- ret :: [a] -> [a] -> [a]
           ret (x:xs) (y:ys) = y `pseq` (x : ret xs ys)
           ret xs     []     = xs

        -- start :: Int -> [a] -> [a]
           start 0   ys     = ys
           start !_n []     = []
           start !n  (y:ys) = y `pseq` start (n-1) ys

-- | 'evalBuffer'' is a rolling buffer strategy combinator for (lazy) lists,
-- much like 'evalBuffer', except that it realises the buffer as a lazy list
-- (with worst-case constant-time operations).
--
-- 'evalBuffer'' is not as compositional as the type suggests. In fact,
-- it evaluates list elements at least to weak head normal form,
-- disregarding a strategy argument 'r0'.
-- 
-- > evalBuffer' n r0 == evalBuffer' n rseq
--
evalBuffer' :: Int -> Strategy a -> Strategy [a]
evalBuffer' n strat = evalFunctor' strat (evalBufferWHNF n)

-- Like evalBufferWHNF but sparks the list elements when pushing them
-- into the buffer.
-- Not to be exported; used in parBuffer' and for optimisation.
parBufferWHNF :: Int -> Strategy [a]
parBufferWHNF n xs = return (ret xs (start n xs))
  where -- ret :: [a] -> [a] -> [a]
           ret (x:xs) (y:ys) = y `par` (x : ret xs ys)
           ret xs     []     = xs

        -- start :: Int -> [a] -> [a]
           start 0   ys     = ys
           start !_n []     = []
           start !n  (y:ys) = y `par` start (n-1) ys

-- | Like 'evalBuffer'' but evaluates the list elements in parallel when
-- pushing them into the buffer.
parBuffer' :: Int -> Strategy a -> Strategy [a]
parBuffer' n strat = evalFunctor' strat (parBufferWHNF n)
-- Alternative definition via evalBuffer' (may compromise firing of RULES):
-- parBuffer' n strat = evalBuffer' n (rpar `dot` strat)

-- The non-compositional buffer strategies might be more efficient than their
-- more general counterparts; use RULES to do the specialisation.

{-# RULES 
"evalBuffer/rseq"  forall n . evalBuffer  n rseq = evalBufferWHNF n
"parBuffer/rseq"   forall n . parBuffer   n rseq = parBufferWHNF  n
"evalBuffer'/rseq" forall n . evalBuffer' n rseq = evalBufferWHNF n
"parBuffer'/rseq"  forall n . parBuffer'  n rseq = parBufferWHNF  n
 #-}


-- --------------------------------------------------------------------------
-- Strategies for tuples

-- | Evaluate the components of a tuple according to the given strategies.
evalTuple2 :: Strategy a -> Strategy b -> Strategy (a,b)
evalTuple2 strat1 strat2 (x1,x2) =
  pure (,) <*> strat1 x1 <*> strat2 x2

evalTuple3 :: Strategy a -> Strategy b -> Strategy c -> Strategy (a,b,c)
evalTuple3 strat1 strat2 strat3 (x1,x2,x3) =
  pure (,,) <*> strat1 x1 <*> strat2 x2 <*> strat3 x3

evalTuple4 :: Strategy a -> Strategy b -> Strategy c -> Strategy d -> Strategy (a,b,c,d)
evalTuple4 strat1 strat2 strat3 strat4 (x1,x2,x3,x4) =
  pure (,,,) <*> strat1 x1 <*> strat2 x2 <*> strat3 x3 <*> strat4 x4

evalTuple5 :: Strategy a -> Strategy b -> Strategy c -> Strategy d -> Strategy e -> Strategy (a,b,c,d,e)
evalTuple5 strat1 strat2 strat3 strat4 strat5 (x1,x2,x3,x4,x5) =
  pure (,,,,) <*> strat1 x1 <*> strat2 x2 <*> strat3 x3 <*> strat4 x4 <*> strat5 x5

evalTuple6 :: Strategy a -> Strategy b -> Strategy c -> Strategy d -> Strategy e -> Strategy f -> Strategy (a,b,c,d,e,f)
evalTuple6 strat1 strat2 strat3 strat4 strat5 strat6 (x1,x2,x3,x4,x5,x6) =
  pure (,,,,,) <*> strat1 x1 <*> strat2 x2 <*> strat3 x3 <*> strat4 x4 <*> strat5 x5 <*> strat6 x6

evalTuple7 :: Strategy a -> Strategy b -> Strategy c -> Strategy d -> Strategy e -> Strategy f -> Strategy g -> Strategy (a,b,c,d,e,f,g)
evalTuple7 strat1 strat2 strat3 strat4 strat5 strat6 strat7 (x1,x2,x3,x4,x5,x6,x7) =
  pure (,,,,,,) <*> strat1 x1 <*> strat2 x2 <*> strat3 x3 <*> strat4 x4 <*> strat5 x5 <*> strat6 x6 <*> strat7 x7

evalTuple8 :: Strategy a -> Strategy b -> Strategy c -> Strategy d -> Strategy e -> Strategy f -> Strategy g -> Strategy h -> Strategy (a,b,c,d,e,f,g,h)
evalTuple8 strat1 strat2 strat3 strat4 strat5 strat6 strat7 strat8 (x1,x2,x3,x4,x5,x6,x7,x8) =
  pure (,,,,,,,) <*> strat1 x1 <*> strat2 x2 <*> strat3 x3 <*> strat4 x4 <*> strat5 x5 <*> strat6 x6 <*> strat7 x7 <*> strat8 x8

evalTuple9 :: Strategy a -> Strategy b -> Strategy c -> Strategy d -> Strategy e -> Strategy f -> Strategy g -> Strategy h -> Strategy i -> Strategy (a,b,c,d,e,f,g,h,i)
evalTuple9 strat1 strat2 strat3 strat4 strat5 strat6 strat7 strat8 strat9 (x1,x2,x3,x4,x5,x6,x7,x8,x9) =
  pure (,,,,,,,,) <*> strat1 x1 <*> strat2 x2 <*> strat3 x3 <*> strat4 x4 <*> strat5 x5 <*> strat6 x6 <*> strat7 x7 <*> strat8 x8 <*> strat9 x9

-- | Evaluate the components of a tuple in parallel according to
-- the given strategies.
parTuple2 :: Strategy a -> Strategy b -> Strategy (a,b)
parTuple2 strat1 strat2 =
  evalTuple2 (rpar `dot` strat1) (rpar `dot` strat2)

parTuple3 :: Strategy a -> Strategy b -> Strategy c -> Strategy (a,b,c)
parTuple3 strat1 strat2 strat3 =
  evalTuple3 (rpar `dot` strat1) (rpar `dot` strat2) (rpar `dot` strat3)

parTuple4 :: Strategy a -> Strategy b -> Strategy c -> Strategy d -> Strategy (a,b,c,d)
parTuple4 strat1 strat2 strat3 strat4 =
  evalTuple4 (rpar `dot` strat1) (rpar `dot` strat2) (rpar `dot` strat3) (rpar `dot` strat4)

parTuple5 :: Strategy a -> Strategy b -> Strategy c -> Strategy d -> Strategy e -> Strategy (a,b,c,d,e)
parTuple5 strat1 strat2 strat3 strat4 strat5 =
  evalTuple5 (rpar `dot` strat1) (rpar `dot` strat2) (rpar `dot` strat3) (rpar `dot` strat4) (rpar `dot` strat5)

parTuple6 :: Strategy a -> Strategy b -> Strategy c -> Strategy d -> Strategy e -> Strategy f -> Strategy (a,b,c,d,e,f)
parTuple6 strat1 strat2 strat3 strat4 strat5 strat6 =
  evalTuple6 (rpar `dot` strat1) (rpar `dot` strat2) (rpar `dot` strat3) (rpar `dot` strat4) (rpar `dot` strat5) (rpar `dot` strat6)

parTuple7 :: Strategy a -> Strategy b -> Strategy c -> Strategy d -> Strategy e -> Strategy f -> Strategy g -> Strategy (a,b,c,d,e,f,g)
parTuple7 strat1 strat2 strat3 strat4 strat5 strat6 strat7 =
  evalTuple7 (rpar `dot` strat1) (rpar `dot` strat2) (rpar `dot` strat3) (rpar `dot` strat4) (rpar `dot` strat5) (rpar `dot` strat6) (rpar `dot` strat7)

parTuple8 :: Strategy a -> Strategy b -> Strategy c -> Strategy d -> Strategy e -> Strategy f -> Strategy g -> Strategy h -> Strategy (a,b,c,d,e,f,g,h)
parTuple8 strat1 strat2 strat3 strat4 strat5 strat6 strat7 strat8 =
  evalTuple8 (rpar `dot` strat1) (rpar `dot` strat2) (rpar `dot` strat3) (rpar `dot` strat4) (rpar `dot` strat5) (rpar `dot` strat6) (rpar `dot` strat7) (rpar `dot` strat8)

parTuple9 :: Strategy a -> Strategy b -> Strategy c -> Strategy d -> Strategy e -> Strategy f -> Strategy g -> Strategy h -> Strategy i -> Strategy (a,b,c,d,e,f,g,h,i)
parTuple9 strat1 strat2 strat3 strat4 strat5 strat6 strat7 strat8 strat9 =
  evalTuple9 (rpar `dot` strat1) (rpar `dot` strat2) (rpar `dot` strat3) (rpar `dot` strat4) (rpar `dot` strat5) (rpar `dot` strat6) (rpar `dot` strat7) (rpar `dot` strat8) (rpar `dot` strat9)


-- --------------------------------------------------------------------------
-- Strategic function application

{-
These are very handy when writing pipeline parallelism asa sequence of
@$@, @$|@ and @$||@'s. There is no need of naming intermediate values
in this case. The separation of algorithm from strategy is achieved by
allowing strategies only as second arguments to @$|@ and @$||@.
-}

-- | Sequential function application. The argument is evaluated using
--   the given strategy before it is given to the function.
($|) :: (a -> b) -> Strategy a -> a -> b
f $| s  = \ x -> let z = x `using` s in z `pseq` f z

-- | Parallel function application. The argument is evaluated using
-- the given strategy, in parallel with the function application.
($||) :: (a -> b) -> Strategy a -> a -> b
f $|| s = \ x -> let z = x `using` s in z `par` f z

-- | Sequential function composition. The result of
-- the second function is evaluated using the given strategy, 
-- and then given to the first function.
(.|) :: (b -> c) -> Strategy b -> (a -> b) -> (a -> c)
(.|) f s g = \ x -> let z = g x `using` s in 
                    z `pseq` f z

-- | Parallel function composition. The result of the second
-- function is evaluated using the given strategy,
-- in parallel with the application of the first function.
(.||) :: (b -> c) -> Strategy b -> (a -> b) -> (a -> c)
(.||) f s g = \ x -> let z = g x `using` s in 
                    z `par` f z

-- | Sequential inverse function composition, 
-- for those who read their programs from left to right.
-- The result of the first function is evaluated using the 
-- given strategy, and then given to the second function.
(-|) :: (a -> b) -> Strategy b -> (b -> c) -> (a -> c)
(-|) f s g = \ x -> let z = f x `using` s in 
                    z `pseq` g z

-- | Parallel inverse function composition,
-- for those who read their programs from left to right.
-- The result of the first function is evaluated using the 
-- given strategy, in parallel with the application of the 
-- second function.
(-||) :: (a -> b) -> Strategy b -> (b -> c) -> (a -> c)
(-||) f s g = \ x -> let z = f x `using` s in 
                    z `par` g z


-- --------------------------------------------------------------------------
-- Old/deprecated stuff

{-# DEPRECATED Done "The Strategy type is now a -> Eval a, not a -> Done" #-}
type Done = ()

{-# DEPRECATED demanding "Use pseq or $| instead" #-}
demanding :: a -> Done -> a
demanding = flip pseq

{-# DEPRECATED sparking "Use par or $|| instead" #-}
sparking :: a -> Done -> a
sparking  = flip par

{-# DEPRECATED (>|) "Use pseq or $| instead" #-}
(>|) :: Done -> Done -> Done 
(>|) = Prelude.seq

{-# DEPRECATED (>||) "Use par or $|| instead" #-}
(>||) :: Done -> Done -> Done 
(>||) = par
